/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - app.state.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {
  HttpClient,
  HttpHeaders,
} from "@angular/common/http";
import {
  Inject,
  Optional,
} from "@angular/core";
import {TranslateService} from "@ngx-translate/core";
import {Navigate} from "@ngxs/router-plugin";
import {
  Actions,
  ofActionErrored,
  StateContext,
  Store,
} from "@ngxs/store";
import * as moment from "moment";
import {Observable} from "rxjs";
import {
  catchError,
  map,
  tap,
} from "rxjs/operators";
import {AppConfigService} from "../../../config/app-config.service";
import {SOLIDIFY_CONSTANTS} from "../../../constants";
import {CookieType} from "../../../cookie-consent/enums/cookie-type.enum";
import {CookieConsentUtil} from "../../../cookie-consent/utils/cookie-consent.util";
import {RegisterDefaultAction} from "../../../decorators/store.decorator";
import {ApiActionNamePartialEnum} from "../../../enums/partial/api-action-name-partial.enum";
import {LocalStoragePartialEnum} from "../../../enums/partial/local-storage-partial.enum";
import {DefaultSolidifyEnvironment} from "../../../environments/environment.solidify-defaults";
import {LocalStorageHelper} from "../../../helpers/local-storage.helper";
import {ApiService} from "../../../http/api.service";
import {APP_OPTIONS} from "../../../injection-tokens/app-options.injection-token";
import {SolidifyHttpErrorResponseModel} from "../../../models/errors/solidify-http-error-response.model";
import {FrontendVersion} from "../../../models/frontend-version.model";
import {ActionSubActionCompletionsWrapper} from "../../../models/stores/base.action";
import {NativeNotificationService} from "../../../services/native-notification.service";
import {NotificationService} from "../../../services/notification.service";
import {
  isFalse,
  isNullOrUndefinedOrWhiteString,
} from "../../../tools/is/is.tool";
import {CacheBustingUtil} from "../../../utils/cache-busting.util";
import {SsrUtil} from "../../../utils/ssr.util";
import {ofSolidifyActionCompleted} from "../../../utils/stores/store.tool";
import {StoreUtil} from "../../../utils/stores/store.util";
import {UserPreferencesUtil} from "../../../utils/user-preferences.util";
import {defaultBaseStateInitValue} from "../base/base.state";
import {BasicState} from "../base/basic.state";
import {SolidifyAppNameSpace} from "./app-namespace.model";
import {SolidifyAppOptions} from "./app-options.model";
import {SolidifyAppStateModel} from "./app-state.model";
import {SolidifyAppAction} from "./app.action";

export const defaultSolidifyAppStateInitValue: () => SolidifyAppStateModel = () =>
  ({
    ...defaultBaseStateInitValue(),
    isTouchInterface: false,
    isApplicationInitialized: false,
    isServerOffline: false,
    isUnableToLoadApp: false,
    isLoggedIn: false,
    isInTourMode: false,
    ignorePreventLeavePopup: false,
    isOpenedSidebarUserGuide: false,
    isOpenedSidebarCookieConsent: false,
    isOpenedBannerCookieConsent: false,
    newCookiesAvailable: undefined,
    appLanguage: undefined,
    theme: undefined,
    darkMode: false,
    preventExit: false,
    preventExitReasonToTranslate: undefined,
    token: undefined,
    frontendVersion: undefined,
  });

// @dynamic
export abstract class SolidifyAppState<TStateModel extends SolidifyAppStateModel> extends BasicState<TStateModel> {
  protected static readonly _AUTHORIZATION_MODULE_NAME: string = "authorization";

  protected readonly _state: TStateModel;
  protected readonly _stateName: string;
  protected readonly _nameSpace: SolidifyAppNameSpace;

  constructor(protected readonly _store: Store,
              protected readonly _translate: TranslateService,
              protected readonly _actions$: Actions,
              protected readonly _apiService: ApiService,
              protected readonly _httpClient: HttpClient,
              protected readonly _notificationService: NotificationService,
              @Optional() @Inject(APP_OPTIONS) protected readonly _optionsState: SolidifyAppOptions,
              protected readonly _environment: DefaultSolidifyEnvironment,
              protected readonly _nativeNotificationService: NativeNotificationService,
              protected readonly _appConfigService: AppConfigService) {
    super();
    this._state = StoreUtil.getStateFromInstance(this);
    this._stateName = StoreUtil.getStateNameFromInstance(this);
    this._nameSpace = _optionsState.nameSpace;
    StoreUtil.initState(SolidifyAppState, this.constructor, this._nameSpace);
    // _environment.appState = _options.baseAppSate;
    // _environment.appNameSpace = _options.nameSpace as SolidifyAppNameSpace;
  }

  // Override if necessary
  protected _isLoggedInRoot(): boolean {
    return false;
  }

  // Override if necessary
  protected _loadModule(): ActionSubActionCompletionsWrapper[] {
    return [];
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.InitApplication)
  initApplication(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.InitApplication): Observable<boolean> {
    return StoreUtil.dispatchSequentialActionAndWaitForSubActionsCompletion(
      ctx,
      [
        {
          action: new SolidifyAppAction.ChangeAppLanguage(UserPreferencesUtil.getPreferredLanguage(this._environment)),
          subActionCompletions: [
            this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.ChangeAppLanguageSuccess)),
          ],
        },
        ...this._loadModule(),
        {
          action: new SolidifyAppAction.Login(),
          subActionCompletions: [
            this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.LoginSuccess)),
            this._actions$.pipe(ofActionErrored(SolidifyAppAction.LoginSuccess)),
            this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.LoginFail)),
          ],
        },
        {
          action: new SolidifyAppAction.InitApplicationParallel(),
          subActionCompletions: [
            this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.InitApplicationParallelSuccess)),
            this._actions$.pipe(ofSolidifyActionCompleted(SolidifyAppAction.InitApplicationParallelFail)),
          ],
        },
      ],
    ).pipe(
      map(result => {
        if (result.success) {
          ctx.dispatch(new SolidifyAppAction.InitApplicationSuccess(action, result));
        } else {
          ctx.dispatch(new SolidifyAppAction.InitApplicationFail(action, result));
        }
        return result.success;
      }),
    );
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.InitApplicationParallel)
  initApplicationParallel(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.InitApplicationParallel): Observable<boolean> {
    return StoreUtil.dispatchParallelActionAndWaitForSubActionsCompletion(
      ctx,
      [
        {action: new SolidifyAppAction.DefineFallbackLanguage()},
        {action: new SolidifyAppAction.SetMomentLocal()},
        ...this._optionsState.initApplicationParallel,
      ],
    ).pipe(
      map(result => {
        if (result.success) {
          ctx.dispatch(new SolidifyAppAction.InitApplicationParallelSuccess(action));
        } else {
          ctx.dispatch(new SolidifyAppAction.InitApplicationParallelFail(action));
        }
        return result.success;
      }),
    );
  }

  // Override if necessary
  protected _updateEnvironmentAllowedUrlsIfNecessary(): void {
    return;
  }

  // Override if necessary
  protected _computeHttpHeadersForLoadBackendModuleVersion(): HttpHeaders {
    return new HttpHeaders();
  }

  protected static _getModuleInfoUrl(name: string, url: string): string {
    return this._getModuleUrl(name, url) + SOLIDIFY_CONSTANTS.URL_SEPARATOR + ApiActionNamePartialEnum.INFO;
  }

  protected static _getModuleInfoDashboard(name: string, url: string): string {
    return this._getModuleUrl(name, url) + SOLIDIFY_CONSTANTS.URL_SEPARATOR + ApiActionNamePartialEnum.DASHBOARD;
  }

  private static _getModuleUrl(name: string, url: string): string {
    if (name !== this._AUTHORIZATION_MODULE_NAME && (url.match(/\//g) || []).length > 2) {
      url = url.substring(0, url.lastIndexOf(SOLIDIFY_CONSTANTS.URL_SEPARATOR));
    }
    return url;
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.InitApplicationSuccess)
  initApplicationSuccess(ctx: StateContext<SolidifyAppStateModel>): void {
    ctx.patchState({
      isApplicationInitialized: true,
    });
    this.subscribe(this._appConfigService.environmentUpdatedObs.pipe(
      tap(environment => {
        this._navigateToMaintenanceIfApplicable();
      }),
    ));
    this._navigateToMaintenanceIfApplicable();
  }

  protected _navigateToMaintenanceIfApplicable(): boolean {
    if (this._environment.maintenanceMode && !this._isLoggedInRoot()) {
      this._store.dispatch(new Navigate([this._optionsState.routeMaintenance]));
      return true;
    }
    return false;
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.InitApplicationFail)
  initApplicationFail(ctx: StateContext<SolidifyAppStateModel>): void {
    if (this._navigateToMaintenanceIfApplicable()) {
      return;
    }
    if (isFalse(ctx.getState().isServerOffline)) {
      ctx.patchState({
        isUnableToLoadApp: true,
      });
      ctx.dispatch(new Navigate([this._optionsState.routeUnableToLoadApp]));
    }
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.DefineFallbackLanguage)
  defineFallbackLanguage(ctx: StateContext<SolidifyAppStateModel>): void {
    // this language will be used as a fallback when a translation isn't found in the current language
    this._translate.setDefaultLang(this._environment.defaultLanguage);
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.SetMomentLocal)
  setMomentLocal(ctx: StateContext<SolidifyAppStateModel>): void {
    const navigatorLanguage = SsrUtil.window?.navigator.language;
    moment.locale(navigatorLanguage);
    // this._adapter.setLocale("fr-CH"); // TOFIX : DOESN'T WORK HERE, NEED TO BE SET IN APP MODULE
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.ChangeAppLanguage)
  changeAppLanguage(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.ChangeAppLanguage): Observable<any> {
    ctx.patchState({
      appLanguage: action.language,
    });

    if (CookieConsentUtil.isFeatureAuthorized(CookieType.localStorage, LocalStoragePartialEnum.language)) {
      LocalStorageHelper.setItem(LocalStoragePartialEnum.language, action.language);
    }
    return this._translate.use(action.language).pipe(
      tap(() => {
        ctx.dispatch(new SolidifyAppAction.ChangeAppLanguageSuccess(action));
      }),
    );
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.ChangeAppLanguageSuccess)
  changeAppLanguageSuccess(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.ChangeAppLanguageSuccess): void {
    // ctx.dispatch(new AppCarouselAction.GetCarousel());
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.ChangeAppTheme)
  changeAppTheme(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.ChangeAppTheme): void {
    this._environment.theme = action.theme;
    ctx.patchState({
      theme: action.theme,
    });
    // Force to refresh language file consolidated with theme json and customizable json
    this._environment.appLanguages.forEach(language => this._translate.reloadLang(language));
    // Force to refresh current page translation by switching of language (only possible if more than one language)
    const currentLanguage = ctx.getState().appLanguage;
    const isFirstLoadApp = isNullOrUndefinedOrWhiteString(currentLanguage);
    if (this._environment.appLanguages.length > 1 && !isFirstLoadApp) {
      const otherLanguage = this._environment.appLanguages.find(l => l !== currentLanguage);
      this.subscribe(StoreUtil.dispatchActionAndWaitForSubActionCompletion(this._store, this._actions$,
        new SolidifyAppAction.ChangeAppLanguage(otherLanguage),
        SolidifyAppAction.ChangeAppLanguageSuccess,
        () => {
          ctx.dispatch(new SolidifyAppAction.ChangeAppLanguage(currentLanguage));
        }));
    }
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.ChangeDarkMode)
  changeDarkMode(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.ChangeDarkMode): void {
    ctx.patchState({
      darkMode: action.enabled,
    });
    this._applyDarkMode(ctx, action);
    if (CookieConsentUtil.isFeatureAuthorized(CookieType.localStorage, LocalStoragePartialEnum.darkMode)) {
      LocalStorageHelper.setItem(LocalStoragePartialEnum.darkMode, action.enabled ? SOLIDIFY_CONSTANTS.STRING_TRUE : SOLIDIFY_CONSTANTS.STRING_FALSE);
    }
  }

  // Override if necessary
  protected _applyDarkMode(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.ChangeDarkMode): void {
    return;
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.ChangeDisplaySidebarUserGuide)
  changeDisplaySidebarUserGuide(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.ChangeDisplaySidebarUserGuide): void {
    ctx.patchState({
      isOpenedSidebarUserGuide: action.isOpen,
    });
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.ChangeDisplayBannerCookieConsent)
  displayCookieConsent(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.ChangeDisplayBannerCookieConsent): void {
    ctx.patchState({
      isOpenedBannerCookieConsent: action.isOpen,
      newCookiesAvailable: action.newCookiesAvailable,
    });
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.ChangeDisplaySidebarCookieConsent)
  changeDisplaySidebarCookieConsent(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.ChangeDisplaySidebarCookieConsent): void {
    ctx.patchState({
      isOpenedSidebarCookieConsent: action.isOpen,
    });
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.Login)
  login(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.Login): ReturnType<this["doLogin"]> {
    return this.doLogin(ctx, action);
  }

  // Override if necessary and type in subclass
  doLogin(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.Login): any {
    return;
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.LoginSuccess)
  loginSuccess(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.LoginSuccess): ReturnType<this["doLoginSuccess"]> {
    return this.doLoginSuccess(ctx, action);
  }

  // Override if necessary and type in subclass
  doLoginSuccess(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.LoginSuccess): any {
    return;
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.LoginFail)
  loginFail(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.LoginFail): ReturnType<this["doLoginFail"]> {
    return this.doLoginFail(ctx, action);
  }

  // Override if necessary and type in subclass
  doLoginFail(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.LoginFail): any {
    return;
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.GetAppVersion)
  getAppVersion(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.GetAppVersion): Observable<FrontendVersion> {
    return this._httpClient.get<FrontendVersion>("./assets/" + this._environment.frontendVersionFile + CacheBustingUtil.generateCacheBustingQueryParam())
      .pipe(
        tap(result => {
          ctx.dispatch(new SolidifyAppAction.GetAppVersionSuccess(action, result));
        }),
        catchError((error: SolidifyHttpErrorResponseModel) => {
          ctx.dispatch(new SolidifyAppAction.GetAppVersionFail(action));
          throw error;
        }),
      );
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.GetAppVersionSuccess)
  getAppVersionSuccess(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.GetAppVersionSuccess): void {
    ctx.patchState({
      frontendVersion: action.appVersion,
    });
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.GetAppVersionFail)
  getAppVersionFail(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.GetAppVersionFail): void {
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.Logout)
  logout(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.Logout): ReturnType<this["doLogout"]> {
    return this.doLogout(ctx, action);
  }

  // Override if necessary and type in subclass
  doLogout(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.Logout): any {
    return;
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.CancelPreventExit)
  cancelPreventExit(ctx: StateContext<SolidifyAppStateModel>): void {
    ctx.patchState({
      preventExit: false,
      preventExitReasonToTranslate: undefined,
    });
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.PreventExit)
  preventExit(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.PreventExit): void {
    ctx.patchState({
      preventExit: true,
      preventExitReasonToTranslate: action.reasonToTranslate,
    });
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.SetIsInTourMode)
  setIsInTourMode(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.SetIsInTourMode): void {
    ctx.patchState({
      isInTourMode: action.isInTourMode,
    });
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.SetIsTouchInterface)
  setIsTouchInterface(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.SetIsTouchInterface): void {
    ctx.patchState({
      isTouchInterface: action.isTouchInterface,
    });
  }

  @RegisterDefaultAction((solidifyAppNameSpace: SolidifyAppNameSpace) => solidifyAppNameSpace.InitiateNewLogin)
  initiateNewLogin(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.InitiateNewLogin): ReturnType<this["doInitiateNewLogin"]> {
    return this.doInitiateNewLogin(ctx, action);
  }

  // Override if necessary and type in subclass
  doInitiateNewLogin(ctx: StateContext<SolidifyAppStateModel>, action: SolidifyAppAction.InitiateNewLogin): any {
    return;
  }

}
